% This function is needed for generating the Drivera database.
%
% Alexandru Paul Condurache, Johannes Kotzerke and Alfred Mertins,
% "Robust retina-based person authentication using the sparse classifier.", 2012.

function retina__ = transformImage(arg, image)
[translation, rotation, scale, shearing, distortion, lowpass, noise, gray, fragments] = parsearg(arg);
dim = size(image);
if numel(dim) == 2
    dim(3) = 1;
end
% Combine all the affine transformations to one matrix.
if isfield(arg, 'matrix')
    T = arg.matrix;
else
    T = [1 0 translation(1); 0 1 translation(2); 0 0 1] ...
        *[scale(1) 0 0; 0 scale(2) 0; 0 0 1] ...
        *[cos(rotation) sin(rotation) 0; -sin(rotation) cos(rotation) 0; 0 0 1] ...
        *[1 shearing(1) 0; shearing(2) 1 0; 0 0 1];
end
% Maketform works [u v] = [ x y 1] T, and not [u v]' = T' [x y 1]' --> T'.
tform = maketform('affine',T');
% Do the transformation for all image layers.
for i = 1:dim(3)
    retina(:,:,i) = imtransform(image(:,:,i), tform, 'fill', 0);
end
dim = size(retina);
retina_ = uint8(zeros(dim(1)+abs(translation(2)), dim(2)+abs(translation(1)), size(retina,3)));
for i = 1:size(retina,3)
    retina_(max([0 translation(2)])+1:dim(1)+max([0 translation(2)]), max([0 translation(1)])+1:dim(2)+max([0 translation(1)]),i) = retina(:,:,i);
end
if distortion ~= 0
    resamp = makeresampler('linear','fill');
    for i = 1:size(retina,3)
        I = retina_(:,:,i);
        [nrows,ncols] = size(I);
        imid = round([nrows ncols] ./2); % Find index of middle element
        [xi,yi] = meshgrid(1:ncols,1:nrows);
        xt = xi(:) - imid(2);
        yt = yi(:) - imid(1);
        [theta,r] = cart2pol(xt,yt);
        s = r + distortion*r.^3;
        [ut,vt] = pol2cart(theta,s);
        u = reshape(ut,size(xi)) + imid(2);
        v = reshape(vt,size(yi)) + imid(1);
        tmap_B = cat(3,u,v);
        retina__(:,:,i) = tformarray(I,[],resamp,[2 1],[1 2],[],tmap_B,.0);
    end
else
    retina__ = retina_;
end
image = retina__;
% Do the non-linear gray changing and filtering after applying the affine
% transformations.
dim = size(image);
if numel(dim) == 2
    dim(3) = 1;
end
if (gray ~= 1) && (gray > 0)
    if dim(3) == 1
        image_ = double(image) ./ 255;
        image = uint8((image_.^gray).*255);
    else
        hsv = rgb2hsv(image);
        hsv(:,:,3) = hsv(:,:,3).^gray;
        image = hsv2rgb(hsv);
        image = uint8(image.*255);
    end
end
% Filtering the image with square lowpass filter with the given size
if lowpass > 1
    filter = generateGaussianFilter(lowpass,2);
    for i = 1:dim(3)
        image(:,:,i) = uint8(conv2(double(image(:,:,i)), filter, 'same'));
    end
end
if sum(sum(noise)) > 0
    % Additive Gaussian noise; variance up to 10 is not a problem.
    if noise(2,1) > 0
        imgfilter = sqrt(noise(2,1))*randn(dim(1:2)) + noise(1,1);
        for i = 1:dim(3)
            image(:,:,i) = uint8(double(image(:,:,i)) + imgfilter);
        end
    % Salt-and-Pepper noise, the given value is the percentage of corrupted
    % pixels. Value should be between 1e-3 and 1e-4.
    elseif (noise(1,2) > 0)
        imgfilter = rand(dim(1:2));
        limit = noise(1,2)/2;
        for i = 1:dim(3)
            img = image(:,:,i);
            img(imgfilter <= limit) = 0;
            img(imgfilter >= (1-limit)) = 255;
            image(:,:,i) = img;
        end
    end
end
if fragments > 0
    imwrite(image, 'tmp.jpg', 'jpeg', 'Mode', 'lossy', 'Quality', fragments);
    image = imread('tmp.jpg');
end
retina__ = image;
end

function [translation, rotation, scale, shearing, distortion, lowpass, noise, gray, fragments] = parsearg(arg)
% Default values.
translation = [0 0];
rotation = 0;
scale = [1 1];
% First value: Horizontal, second vertical shearing.
shearing = [0 0];
distortion = 0;
lowpass = 1;
noise = [0 0; 0 0];
gray = 1;
% Parse arguments from input.
if isfield(arg, 'translation')
    translation = arg.translation;
    if length(translation) == 1
        % interpretation as horizontal translation
        translation = [translation 0];
    end
end
if isfield(arg, 'rotation')
    rotation = arg.rotation*pi/180;
end
if isfield(arg, 'scale')
    scale = arg.scale;
    if length(scale) == 1
        scale = [scale scale];
    end
end
if isfield(arg, 'shearing')
    shearing = arg.shearing;
    if length(shearing) == 1
        % interpretation as horizontal shearing
        shearing = [shearing 0];
    end
end
if isfield(arg, 'distortion')
    distortion = arg.distortion;
end
if isfield(arg, 'lowpass')
    lowpass = arg.lowpass;
end
if isfield(arg, 'noise')
    % the first value stands for gaussian noise, the second for
    % salt-and-pepper noise
    noise = arg.noise;
end
if isfield(arg, 'gray')
    gray = arg.gray;
end
if isfield(arg, 'fragments')
    fragments = arg.fragments;
end
end

function filter = generateGaussianFilter(width, d)
if width == 1
    filter = 1;
else
    filter = [1 1];
    for i = 1:(width-2)
        filter = conv(filter, [1 1]);
    end
    if d == 2
        filter = filter'*filter;
    end
end
filter = filter ./ sum(sum(filter));
end